//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/RealItem.h
//! @brief     Defines class RealItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DATA_REALITEM_H
#define BORNAGAIN_GUI_MODEL_DATA_REALITEM_H

#include "GUI/Model/Data/DataItem.h"
#include "GUI/Support/IO/AbstractDataLoader.h"

class Datafield;
class ImportDataInfo;
class InstrumentItem;
class IntensityDataItem;
class MaskContainerItem;
class MessageService;
class SpecularDataItem;

//! Provides access to experimental data, for display and fitting.
//! Owns an AbstractDataLoader.

class BA_CORE_API_ RealItem : public QObject {
    Q_OBJECT

    friend class TestView;

public:
    RealItem();

    //! The name which is presented to the user
    QString realItemName() const;
    void setRealItemName(const QString& name);

    QString presentationType() const;
    void setPresentationType(const QString& type);

    // data

    DataItem* dataItem() const;
    IntensityDataItem* intensityDataItem() const;
    SpecularDataItem* specularDataItem() const;

    DataItem* initFromType(const QString& type);
    void initAsSpecularItem();
    void initAsIntensityItem();

    bool isIntensityData() const;
    bool isSpecularData() const;

    // native data

    bool hasNativeData() const;
    DataItem* nativeDataItem() const;

    DataItem* initNativeData();
    void removeNativeData();

    QString nativeDataUnits() const;
    void setNativeDataUnits(const QString& units);
    bool holdsDimensionalData() const;

    const Datafield* nativeDatafield() const;
    void setNativeDatafield(Datafield* data);

    void setDatafield(Datafield* data);
    void setImportData(ImportDataInfo info);

    //! The name from where the native data was originally imported
    void setNativeFileName(const QString& filename);
    QString nativeFileName() const;

    // instrument

    QString instrumentId() const;
    void linkToInstrument(const InstrumentItem* instrument);
    void unlinkFromInstrument();

    // loader

    bool hasImportErrors() const;

    //! Takes ownership of loader
    void setDataLoader(AbstractDataLoader* loader);
    AbstractDataLoader* dataLoader() const;

    //! Returns true, if rotation will affect linked instrument or mask presence.
    bool rotationAffectsSetup() const;
    void rotateData();

    void updateDataFileName();

    // serialization

    QByteArray serializeBinaryData() const;
    void deserializeBinaryData(const QByteArray& data);

    void writeTo(QXmlStreamWriter* w, const QString& projectDir = "") const;
    QString readFrom(QXmlStreamReader* r, const QString& projectDir = "",
                     MessageService* messageService = nullptr);

    // other

    //! Make a copy instance.
    void copyTo(RealItem* const realdata_new) const;

    //! Returns the shape of underlying data item
    std::vector<int> shape() const;

    //! Returns mask container item
    MaskContainerItem* maskContainerItem();

signals:
    void importContentsProcessed();

private:
    void initDataItem(const std::size_t data_rank, std::unique_ptr<DataItem>& dataItem);

    //! This does not set the link to the instrument! Use linkToInstrument() if you want
    //! to link to this instrument.
    void updateToInstrument(const InstrumentItem* instrument);

private:
    QString m_nativeFileName;
    QString m_instrumentId;
    QString m_name;
    QString m_nativeDataUnits = "nbins";
    QString m_presentationType;

    // DataItem can be `IntensityDataItem` and `SpecularDataItem` (default `IntensityDataItem`).
    std::unique_ptr<DataItem> m_dataItem;       // TODO: Can be replaced by std::variant
    std::unique_ptr<DataItem> m_nativeDataItem; // TODO: Can be replaced by std::variant
    std::unique_ptr<AbstractDataLoader> m_dataLoader;
};

#endif // BORNAGAIN_GUI_MODEL_DATA_REALITEM_H
