//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/PlotUtil/RangeUtils.cpp
//! @brief     Implements RangeUtils namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/PlotUtil/RangeUtils.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Util/Error.h"
#include "GUI/View/Tool/StyleUtils.h"
#include <qcustomplot.h>

namespace {

// Converts xmin (low edge of first bin) and xmax (upper edge of last bin) to the
// range expected by QCPColorMapData::setRange.
QCPRange qcpRange(double xmin, double xmax, int nbins)
{
    double dx = (xmax - xmin) / nbins;
    return QCPRange(xmin + dx / 2., xmax - dx / 2.);
}

QMargins defaultMargins(const QWidget& widget)
{
    auto base_size = GUI::Util::Style::SizeOfLetterM(&widget);
    int left = static_cast<int>(base_size.width() * 6.0);
    int top = static_cast<int>(base_size.height() * 1.5);
    int right = static_cast<int>(base_size.width() * 1.2);
    int bottom = static_cast<int>(base_size.height() * 4.5);
    return QMargins(left, top, right, bottom);
}

} // namespace

QCPRange GUI::View::RangeUtils::itemXrange(const IntensityDataItem* item)
{
    return qcpRange(item->xMin(), item->xMax(), item->xSize());
}

QCPRange GUI::View::RangeUtils::itemZoomX(const IntensityDataItem* item)
{
    return QCPRange(item->lowerX(), item->upperX());
}

QCPRange GUI::View::RangeUtils::itemYrange(const IntensityDataItem* item)
{
    return qcpRange(item->yMin(), item->yMax(), item->ySize());
}

QCPRange GUI::View::RangeUtils::itemZoomY(const IntensityDataItem* item)
{
    return QCPRange(item->lowerY(), item->upperY());
}

QCPRange GUI::View::RangeUtils::itemDataRange(const IntensityDataItem* item)
{
    QPair<double, double> range = item->dataRange();
    return QCPRange(range.first, range.second);
}

QCPRange GUI::View::RangeUtils::itemDataZoom(const IntensityDataItem* item)
{
    return QCPRange(item->lowerZ(), item->upperZ());
}

void GUI::View::RangeUtils::setLogz(QCPColorScale* scale, bool isLogz)
{
    if (isLogz && scale->dataScaleType() != QCPAxis::stLogarithmic)
        scale->setDataScaleType(QCPAxis::stLogarithmic);

    else if (!isLogz && scale->dataScaleType() != QCPAxis::stLinear)
        scale->setDataScaleType(QCPAxis::stLinear);

    setLogz(scale->axis(), isLogz);
}

void GUI::View::RangeUtils::setLogz(QCPAxis* axis, bool isLogz)
{
    if (isLogz) {
        axis->setNumberFormat("eb");
        axis->setNumberPrecision(0);
        axis->setScaleType(QCPAxis::stLogarithmic);
        QSharedPointer<QCPAxisTicker> ticker(new QCPAxisTickerLog);
        axis->setTicker(ticker);
    } else {
        axis->setNumberFormat("f");
        axis->setNumberPrecision(1);
        axis->setScaleType(QCPAxis::stLinear);
        QSharedPointer<QCPAxisTicker> ticker(new QCPAxisTicker);
        axis->setTicker(ticker);
    }
}

void GUI::View::RangeUtils::setDefaultMargins(QCustomPlot* customPlot)
{
    auto* axisRectangle = customPlot->axisRect();
    axisRectangle->setAutoMargins(QCP::msTop | QCP::msBottom);
    axisRectangle->setMargins(defaultMargins(*customPlot));
}

double GUI::View::RangeUtils::commonMin(const QList<AmplitudeAxisItem*>& axes)
{
    double min = +std::numeric_limits<double>::max();
    for (auto axis : axes)
        if (min > axis->min())
            min = axis->min();
    return min;
}

double GUI::View::RangeUtils::commonMax(const QList<AmplitudeAxisItem*>& axes)
{
    double max = -std::numeric_limits<double>::max();
    for (auto axis : axes)
        if (max < axis->max())
            max = axis->max();
    return max;
}

QCPRange GUI::View::RangeUtils::commonRange(const QList<AmplitudeAxisItem*>& axes)
{
    return QCPRange(commonMin(axes), commonMax(axes));
}

void GUI::View::RangeUtils::setCommonRangeY(QList<SpecularDataItem*> items)
{
    QCPRange range = commonRange(valueAxesFromSpecularItems(items));
    for (auto item : items) {
        item->setLowerY(range.lower);
        item->setUpperY(range.upper);
    }
}

void GUI::View::RangeUtils::setCommonRangeZ(QList<IntensityDataItem*> items)
{
    QCPRange range = commonRange(valueAxesFromIntensityItems(items));
    for (auto item : items) {
        item->setLowerZ(range.lower);
        item->setUpperZ(range.upper);
    }
}

QList<AmplitudeAxisItem*>
GUI::View::RangeUtils::valueAxesFromSpecularItems(const QList<SpecularDataItem*>& items)
{
    QList<AmplitudeAxisItem*> axes;
    for (auto item : items)
        axes.append(item->yAxisItem());
    return axes;
}

QList<AmplitudeAxisItem*>
GUI::View::RangeUtils::valueAxesFromIntensityItems(const QList<IntensityDataItem*>& items)
{
    QList<AmplitudeAxisItem*> axes;
    for (auto item : items)
        axes.append(item->zAxisItem());
    return axes;
}
